//			combine.C
//
//		transform_combined()
//
// written by Zheng Gengbin

/*
This pass transform combined directive to several directives equally .
first delete the dimension_op , and add the dimension info directly to
the varibable list .
split the combined directive to sevaral directives by variable list according to 
every attribute .
As to align_op ,if align subscript is omitted , add it back according to the
dimension of the array .
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "compilerApi.h"
#include "pass.h"


//  add the dim in DIMENSION_OP to the var list
void	adjust_var_list(SgStatement *s,SgExpression *dim)
{
	SgExpression *var_list,*ele;

	for (var_list=s->expr(0) ; var_list; var_list=var_list->rhs()) 
	{
		ele=var_list->lhs();
		if (ele->variant() != ARRAY_REF) 
		{
			ele->thellnd->variant=ARRAY_REF;
			ele->setLhs(dim);
		}
	}
}

void create_proc_or_temp__stmt(int variant,SgStatement *s)
{
	SgStatement *decl;

	//decl=new SgVarDeclStmt(* s->expr(0),* s->expr(1)->type());
	SgExpression *type_op = new SgExpression(TYPE_OP,NULL,NULL,NULL,s->expr(1)->type());
	decl=new SgStatement(variant,NULL,NULL,s->expr(0),type_op,NULL);
	//decl->setVariant(variant);
	s->insertStmtBefore(*decl,* s->controlParent() );
}

void create_dist_or_align_stmts(int variant,SgStatement *s,SgExpression *attr)
{
	SgStatement *decl;
	SgExpression *var_list,*var,*op,*att;
	SgExpression *src_list=NULL,*spec_list=NULL;

	//op=s->expr(2)->lhs()->lhs();       //point to DISTRIBUTE_OP or ALIGN_OP
	op=attr->lhs();

	// fill in the align_source_list if missed !!
	if (op->variant() == ALIGN_OP)
	{
		SgSymbol *align_target=op->rhs()->rhs()->symbol();
		SgExpression *align_source = op->lhs();
		SgExpression *align_spec = op->rhs()->rhs()->lhs();

		if (( align_source == NULL || align_spec == NULL ) && isSgArrayType(align_target->type()) )
		{
			for (int i = 0; i<((SgArrayType*)(align_target->type()))->dimension() ; i++)
			{
			   SgExpression *node;
			   if (align_spec == NULL ) {
				node=new SgExpression (DDOT,NULL,NULL,NULL);
				spec_list=new SgExpression(EXPR_LIST,node,spec_list,NULL);
                           }
			   if (align_source == NULL ) {
				node=new SgExpression (COLON_OP,NULL,NULL,NULL);
				src_list=new SgExpression(EXPR_LIST,node,src_list,NULL);
			   }
			}
		}
	}

	for (var_list=s->expr(0) ; var_list; var_list=var_list->rhs()) 
	{
		SgExpression *ele = var_list->lhs();
		SgExpression *srcList = src_list;
		  // if varRef already have dummy declared here, we donot use the
		  //	colon_op created above(src_list)
		if ( ele->variant() == ARRAY_REF ) {
			SgArrayRefExp *array = isSgArrayRefExp(ele);
			ele = new SgVarRefExp(ele->symbol());
			srcList = NULL;
			for (int i=0; i<array->numberOfSubscripts(); i++) {
				SgExpression  *sub = array->subscript(i);
				SgSubscriptExp *ddot = isSgSubscriptExp(sub);
				if (ddot) sub = ddot->ubound();
				srcList=new SgExpression(EXPR_LIST,sub->copyPtr(),srcList,NULL);
			}
		}
		else
			ele = ele->copyPtr();
		var=new SgExpression(EXPR_LIST,ele,NULL,NULL);
		SgExpression *new_op = op->copyPtr();
		  // add  align source or target list if anyone is missed
		if ( srcList )
			new_op->setLhs(srcList);
		if ( spec_list )
		{
			new_op->rhs()->rhs()->setLhs(spec_list);
			new_op->rhs()->rhs()->thellnd->variant = ARRAY_REF;
		}
		att=new SgExpression(EXPR_LIST , new_op,NULL,NULL);
		decl=new SgStatement(variant,NULL,NULL,var, s->expr(1),att);
		s->insertStmtBefore(*decl,* s->controlParent() );
	}
}

void create_inherit_dynamic_stmts(int variant,SgStatement *s,SgExpression *attr)
{
	SgStatement *decl=new SgStatement(variant,NULL,NULL,s->expr(0),NULL,attr);
	//decl->setVariant(variant);
	s->insertStmtBefore(*decl,* s->controlParent() );
}

void transform_combined(SgStatement *s)
{
	SgExpression  *var_list,*attr_list,*attr;
	SgStatement   *nstmt;

	var_list=s->expr(0);
	attr_list=s->expr(2);

	// deal with dimension first
	for (attr_list=s->expr(2) ; attr_list ; attr_list=attr_list->rhs())
	{
		attr=attr_list->lhs();
		if (attr->variant() == DIMENSION_OP )
		{
			adjust_var_list(s,attr->lhs());
		}
	}

	for (attr_list=s->expr(2) ; attr_list ; attr_list=attr_list->rhs())
	{
		attr=attr_list->lhs();
		switch (attr->variant())
		{
		case PROCESSORS_VAL:
			create_proc_or_temp__stmt(PROCESSORS_STMT,s);
			break;
		case TEMPLATE_VAL:
			create_proc_or_temp__stmt(TEMPLATE_STMT,s);
			break;
		case DISTRIBUTE_ATT:
			create_dist_or_align_stmts(DISTRIBUTE_DECL,s,attr);
			break;
		case ALIGN_ATT:
			create_dist_or_align_stmts(ALIGN_DECL,s,attr);
			break;
		case INHERIT_OP:
			create_inherit_dynamic_stmts(INHERIT_DECL,s,attr);
			break;
		case DYNAMIC_OP:
			create_inherit_dynamic_stmts(DYNAMIC_DECL,s,attr);
			break;
		}
	}

	// delete this statement
	s->deleteStmt();
}
